from abc import ABC
from collections.abc import Iterator, Mapping
from enum import Enum, unique
from typing import Generic, TypeGuard, TypeVar

from .group import Group

_G = TypeVar("_G", bound=Group)


@unique
class Page(Enum):
    ORIGINAL = 0
    FRIEND = 1


class Save(ABC, Generic[_G], Mapping[Page, _G]):
    __slots__ = ()

    def __contains__(self, key: object, /) -> TypeGuard[Page]:
        return isinstance(key, Page)

    def __iter__(self) -> Iterator[Page]:
        return iter(Page)

    def __len__(self) -> int:
        return 2



from abc import ABC
from collections.abc import Iterable, Sequence
from itertools import filterfalse
from typing import Final, Generic, Optional, TypeVar

from .level import Level, LevelNotFoundError

_L = TypeVar("_L", bound=Level)


class Group(ABC, Generic[_L], Sequence[_L]):
    """A group of 20 levels.
    Note that levels are 0-indexed within this interface, but 1-indexed in-game.
    """

    __slots__ = ()

    def fill_mask(self) -> Sequence[bool]:
        """Return which level IDs within this set exist."""
        return [bool(l) for l in self]

    def __len__(self) -> int:
        return 20

    def read(self) -> Iterable[Optional[bytes]]:
        """Read all of the levels within this Group. Empty slots yield None."""
        content: Optional[bytes]
        result: Final[list[Optional[bytes]]] = []
        for level in self:
            try:
                content = level.read()
            except LevelNotFoundError:
                content = None
            result.append(content)
        return result

    def write(self, new_content: Iterable[Optional[bytes]], /) -> None:
        """Replace the contents of this Group with the specified new content. None values will empty the slot they correspond to."""
        for src, dest in filterfalse(
            lambda x: x[0] is None, zip(new_content, self, strict=True)
        ):
            dest.write(src)



from abc import ABC, abstractmethod
from collections.abc import Sized
from dataclasses import dataclass
from enum import Enum, unique
from typing import Final


@unique
class Theme(Enum):
    THE_EMPTY_LOT = 0
    NEIGHBORS_HOUSE = 1
    SIZZLIN_DESERT = 2
    CHILL_MOUNTAIN = 3
    OCEAN_TREASURE = 4
    SPACE_STATION = 5
    STUMP_TEMPLE = 6
    CANDY_ISLAND = 7
    HAUNTED_HOUSE = 8
    CITY = 9
    TUTORIAL = 11
    HAUNTED_HOUSE_DARKNESS = 12
    NIGHT_CITY = 13

    def __str__(self) -> str:
        return (
            "The Empty Lot",
            "Neighbor's House",
            "Sizzlin' Desert",
            "Chill Mountain",
            "Ocean Treasure",
            "Space Station",
            "Stump Temple",
            "Candy Island",
            "Haunted House",
            "City",
            None,
            "Tutorial",
            "Haunted House Darkness",
            "Night City",
        )[self.value]


@dataclass(frozen=True, match_args=False, kw_only=True, slots=True)
class LevelStatistics:
    crystals: int
    filesize: int
    theme: Theme


class LevelNotFoundError(LookupError):
    pass


class Level(ABC, Sized):
    __slots__ = ()

    def about(self) -> LevelStatistics:
        content: Final[bytes] = self.read()
        return LevelStatistics(
            crystals=content.count(b"<anmtype> 49 </anmtype>"),
            filesize=len(content),
            theme=Theme(int(content[87:89])),
        )

    @abstractmethod
    def __bool__(self) -> bool:
        """Return whether this level exists."""

    @abstractmethod
    def delete(self) -> None:
        """Delete this level if it exists, otherwise raise LevelNotFoundError."""
        pass

    def encode(self) -> bytes:
        """Return a bytes object that when written to a file can overwrite an official level."""
        data: Final[bytearray] = bytearray(self.read())
        header: Final[bytes] = (
            b"\x00\x00\x00\x01\x00\x00\x00\x08"
            + len(data).to_bytes(4, "big")
            + b"\x00\x00\x00\x01"
        )
        i: int = 0
        while i < len(data):
            data.insert(i, 255)
            i += 9
        return header + data + b"\x00"

    def __len__(self) -> int:
        """The file size of this level."""
        return len(self.read())

    @abstractmethod
    def read(self) -> bytes:
        """Return the contents of this level if it exists, otherwise raise LevelNotFoundError."""
        pass

    @abstractmethod
    def write(self, new_content: bytes, /) -> None:
        """Replace the contents of this level, or create it if it doesn't exist."""
        pass



from collections.abc import Sequence
from operator import index
from os import SEEK_CUR
from os.path import abspath, dirname, isfile, join
from typing import Final, Optional, SupportsIndex, TypeGuard, overload

from ..item.group import Group
from ..item.level import Level, LevelNotFoundError
from ..item.save import Page, Save

from . import Location

_BLOCK_SIZE: Final[int] = 2048
_EMPTY_BLOCK: Final[bytes] = b"\x00" * _BLOCK_SIZE
_LEVEL_ALLOCATION_SIZE: Final[int] = 156864


class DirLevel(Level):
    __match_args__ = ("path", "page", "id")
    __slots__ = ("_offset", "_path")

    _offset: int
    _path: str

    def __init__(self, path: str, /, page: Page, id: SupportsIndex) -> None:
        i = index(id)
        if 0 <= i < 20:
            self._offset = 8 + _LEVEL_ALLOCATION_SIZE * (i & 3)
            self._path = join(abspath(path), f"ed0{(id >> 2) + 5 * page.value}.dat")
        else:
            raise ValueError(id)

    def __bool__(self) -> bool:
        with open(self._path, "rb") as f:
            f.seek(self._offset)
            return f.read(1) != b"\x00"

    def delete(self) -> None:
        with open(self._path, "rb") as f:
            f.seek(self._offset)
            more: bool = f.read(1) != b"\x00"
            if more:
                while more:
                    f.seek(-1, SEEK_CUR)
                    f.write(_EMPTY_BLOCK)
                    more = f.read(1) != b"\x00"
            else:
                raise LevelNotFoundError

    def __eq__(self, other: object, /) -> bool:
        return (
            self._offset == other._offset and self._path == other._path
            if isinstance(other, DirLevel)
            else NotImplemented
        )

    def __hash__(self) -> int:
        return hash((self._offset, self._path))

    @property
    def id(self) -> int:
        return int(self._path[-5]) % 5 << 2 | self._offset // _LEVEL_ALLOCATION_SIZE

    def __len__(self) -> int:
        with open(self._path, "rb") as f:
            f.seek(self._offset)
            block: Final[bytearray] = bytearray(f.read(_BLOCK_SIZE))
            block_offset: int = 0
            while block[-1]:
                f.readinto(block)
                block_offset += _BLOCK_SIZE
            hi: int = _BLOCK_SIZE - 1
            lo: int = 0
            test: int
            while hi != lo:
                test = (hi - lo >> 1) + lo
                if block[test]:
                    lo = test + 1
                else:
                    hi = test
            return block_offset + hi

    @property
    def page(self) -> Page:
        return Page(ord(self._path[-5]) > 52)

    @property
    def path(self) -> str:
        return dirname(self._path)

    def read(self) -> bytes:
        with open(self._path, "rb") as f:
            f.seek(self._offset)
            block: Final[bytearray] = bytearray(f.read(_BLOCK_SIZE))
            if block[0]:
                result: Final[bytearray] = bytearray()
                while block[-1]:
                    result.extend(block)
                    f.readinto(block)
                hi: int = _BLOCK_SIZE - 1
                lo: int = 0
                test: int
                while hi != lo:
                    test = (hi - lo >> 1) + lo
                    if block[test]:
                        lo = test + 1
                    else:
                        hi = test
                return bytes(result + block[:hi])
            else:
                raise LevelNotFoundError

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self.path!r}, {self.page!r}, {self.id!r})"

    def write(self, new_content: bytes, /) -> None:
        with open(self._path, "r+b") as f:
            f.seek(self._offset)
            f.write(new_content)
            while f.read(1) != b"\x00":
                f.seek(-1, SEEK_CUR)
                f.write(_EMPTY_BLOCK)


class DirGroup(Group[DirLevel]):
    __match_args__ = ("path", "page")
    __slots__ = ("_page", "_path")

    _page: Page
    _path: str

    def __init__(self, path: str, /, page: Page) -> None:
        self._page = page
        self._path = abspath(path)

    def __contains__(self, value: object, /) -> TypeGuard[DirLevel]:
        return (
            isinstance(value, DirLevel)
            and value.path == self.path
            and value.page is self.page
        )

    def count(self, value: object) -> int:
        return int(value in self)

    def __eq__(self, other: object, /) -> bool:
        return (
            self.page == other.page and self.path == other.path
            if isinstance(other, DirGroup)
            else NotImplemented
        )

    @overload
    def __getitem__(self, index: SupportsIndex, /) -> DirLevel:
        pass

    @overload
    def __getitem__(self, indices: slice, /) -> Sequence[DirLevel]:
        pass

    def __getitem__(
        self, key: SupportsIndex | slice, /
    ) -> DirLevel | Sequence[DirLevel]:
        if isinstance(key, slice):
            return tuple((self[i] for i in range(*key.indices(20))))
        else:
            i: int = index(key)
            if -20 <= i < 20:
                return DirLevel(self.path, self.page, i % 20)
            else:
                raise IndexError(key)

    def __hash__(self) -> int:
        return hash((self.path, self.page))

    def index(self, value: object, start: int = 0, stop: Optional[int] = None) -> int:
        if value in self and value.id in range(*slice(start, stop).indices(20)):
            return value.id
        else:
            raise ValueError

    @property
    def page(self) -> Page:
        return self._page

    @property
    def path(self) -> str:
        return self._path

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self.path!r}, {self.page!r})"


class DirSave(Location, Save[DirGroup]):
    __slots__ = ()

    def __bool__(self) -> bool:
        return all((isfile(join(self.path, f"ed0{i}.dat")) for i in range(10)))

    def __getitem__(self, key: Page, /) -> DirGroup:
        if isinstance(key, Page):
            return DirGroup(self.path, key)
        else:
            raise KeyError(key)



from collections.abc import Iterable, Sequence
from itertools import filterfalse
from operator import index
from os.path import abspath
from re import fullmatch
from typing import Final, Optional, SupportsIndex, TypeGuard, overload
from zipfile import ZipFile, ZipInfo

from ..item.group import Group
from ..item.level import Level, LevelNotFoundError

from . import Location

__all__ = ["ZipGroup", "ZipLevel"]


def _id_to_fn(id: SupportsIndex, /) -> str:
    return f"{str(index(id) + 1).zfill(2)}.lvl"


class ZipLevelNotFoundError(KeyError, LevelNotFoundError):
    pass


class ZipLevel(Level):
    __match_args__ = ("path", "id")
    __slots__ = ("_id", "_path")

    _id: int
    _path: str

    def __init__(self, path: str, /, id: SupportsIndex) -> None:
        i: int = index(id)
        if 0 <= i < 20:
            self._id = i
            self._path = abspath(path)
        else:
            raise ValueError(id)

    def __bool__(self) -> bool:
        with ZipFile(self.path) as a:
            return self.fn in a.namelist()

    def delete(self) -> None:
        with ZipFile(self.path) as a:
            if self.fn not in a.namelist():
                raise ZipLevelNotFoundError
            contents: Final[dict[ZipInfo, bytes]] = {}
            for info in filterfalse(
                lambda info: info.filename == self.fn, a.infolist()
            ):
                contents[info] = a.read(info)
        with ZipFile(self.path, "w") as a:
            for x in contents.items():
                a.writestr(*x)

    def __eq__(self, other: object, /) -> bool:
        return (
            self.path == other.path and self.id == other.id
            if isinstance(other, ZipLevel)
            else NotImplemented
        )

    def __hash__(self) -> int:
        return hash((self.path, self.id))

    @property
    def fn(self) -> str:
        return _id_to_fn(self.id)

    @property
    def id(self) -> int:
        return self._id

    def __len__(self) -> int:
        try:
            with ZipFile(self.path) as a:
                return a.getinfo(self.fn).file_size
        except KeyError as e:
            raise ZipLevelNotFoundError(*e.args)

    @property
    def path(self) -> str:
        return self._path

    def read(self) -> bytes:
        try:
            with ZipFile(self.path) as a:
                return a.read(self.fn)
        except KeyError as e:
            raise ZipLevelNotFoundError(*e.args)

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self.path!r}, {self.id!r})"

    def write(self, new_content: bytes, /) -> None:
        contents: Final[dict[ZipInfo, bytes]] = {}
        with ZipFile(self.path) as a:
            for info in filterfalse(
                lambda info: info.filename == self.fn, a.infolist()
            ):
                contents[info] = a.read(info)
        with ZipFile(self.path, "w") as a:
            for x in contents.items():
                a.writestr(*x)
            a.writestr(self.fn, new_content)


class ZipGroup(Group[ZipLevel], Location):
    __slots__ = ()

    def __contains__(self, value: object, /) -> TypeGuard[ZipLevel]:
        return isinstance(value, ZipLevel) and value.path == self.path

    def count(self, value: object) -> int:
        return int(value in self)

    @overload
    def __getitem__(self, index: SupportsIndex, /) -> ZipLevel:
        pass

    @overload
    def __getitem__(self, indices: slice, /) -> Sequence[ZipLevel]:
        pass

    def __getitem__(
        self, key: SupportsIndex | slice, /
    ) -> ZipLevel | Sequence[ZipLevel]:
        if isinstance(key, slice):
            return tuple((self[i] for i in range(*key.indices(20))))
        elif -20 <= index(key) < 20:
            return ZipLevel(self.path, index(key) % 20)
        else:
            raise IndexError(key)

    def index(self, value: object, start: int = 0, stop: Optional[int] = None) -> int:
        if value in self and value.id in range(*slice(start, stop).indices(20)):
            return value.id
        else:
            raise ValueError

    def init(self) -> None:
        with ZipFile(self.path, "x") as a:
            pass

    def fill_mask(self) -> Sequence[bool]:
        with ZipFile(self.path) as a:
            return [_id_to_fn(i) in a.namelist() for i in range(20)]

    def read(self) -> Iterable[Optional[bytes]]:
        with ZipFile(self.path) as a:
            name_list: Final[list[str]] = a.namelist()
            return [
                a.read(fn) if fn in name_list else None
                for fn in map(_id_to_fn, range(20))
            ]

    def write(self, new_content: Iterable[Optional[bytes]], /) -> None:
        contents: Final[dict[ZipInfo, bytes]] = {}
        if self:
            with ZipFile(self.path) as a:
                for info in filterfalse(
                    lambda info: fullmatch(r"(0[1-9]|1\d|20)\.lvl", info.filename),
                    a.infolist(),
                ):
                    contents[info] = a.read(info)
        with ZipFile(self.path, "w") as a:
            for x in contents.items():
                a.writestr(*x)
            for id, content in filterfalse(
                lambda x: x[1] is None, enumerate(new_content)
            ):
                a.writestr(_id_to_fn(id), content)



from os.path import getsize

from ..item.level import Level, LevelNotFoundError

from . import Location

__all__ = ["LvlLevel"]


class LvlLevelNotFoundError(FileNotFoundError, LevelNotFoundError):
    pass


class LvlLevel(Location, Level):
    __slots__ = ()

    def delete(self) -> None:
        try:
            return super().delete()
        except FileNotFoundError as e:
            raise LvlLevelNotFoundError(*e.args)

    def __len__(self) -> int:
        return getsize(self.path)

    def read(self) -> bytes:
        with open(self.path, "rb") as f:
            return f.read()

    def write(self, new_content: bytes, /) -> None:
        with open(self.path, "wb") as f:
            f.write(new_content)



from os import remove
from os.path import abspath, exists
from shutil import rmtree


class Location:
    __match_args__ = ("path",)
    __slots__ = ("_path",)

    _path: str

    def __init__(self, path: str, /) -> None:
        self._path = abspath(path)

    def __bool__(self) -> bool:
        """Whether this file location is accessible and exists."""
        return exists(self.path)

    def delete(self) -> None:
        try:
            remove(self.path)
        except IsADirectoryError:
            rmtree(self.path)

    def __eq__(self, other: object, /) -> bool:
        return (
            self.path == other.path if isinstance(other, Location) else NotImplemented
        )

    def __hash__(self) -> int:
        return hash(self.path)

    @property
    def path(self) -> str:
        return self._path

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self.path!r})"
